use std::collections::BTreeSet;
use {DataType, ErrorKind, HasDataType, Result};

pub trait IsParameterMask: Clone {}

impl IsParameterMask for BTreeSet<String> {}
impl IsParameterMask for u64 {}

pub type DataValueRaw = DataValue<u64>;
pub type DataValueDescriptive = DataValue<BTreeSet<String>>;

#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
#[cfg_attr(feature = "serde", serde(rename_all = "lowercase", untagged))]
pub enum DataValue<M: IsParameterMask> {
    Boolean(bool),
    UInt8(u8),
    UInt16(u16),
    UInt32(u32),
    UInt64(u64),
    Int8(i8),
    Int16(i16),
    Int32(i32),
    Int64(i64),
    ParameterMask(M),
}

impl<T: IsParameterMask> From<bool> for DataValue<T> {
    fn from(v: bool) -> DataValue<T> {
        DataValue::Boolean(v)
    }
}

impl<T: IsParameterMask> From<i8> for DataValue<T> {
    fn from(v: i8) -> DataValue<T> {
        DataValue::Int8(v)
    }
}

impl<T: IsParameterMask> From<i16> for DataValue<T> {
    fn from(v: i16) -> DataValue<T> {
        DataValue::Int16(v)
    }
}

impl<T: IsParameterMask> From<i32> for DataValue<T> {
    fn from(v: i32) -> DataValue<T> {
        DataValue::Int32(v)
    }
}

impl<T: IsParameterMask> From<i64> for DataValue<T> {
    fn from(v: i64) -> DataValue<T> {
        DataValue::Int64(v)
    }
}

impl<T: IsParameterMask> From<u8> for DataValue<T> {
    fn from(v: u8) -> DataValue<T> {
        DataValue::UInt8(v)
    }
}

impl<T: IsParameterMask> From<u16> for DataValue<T> {
    fn from(v: u16) -> DataValue<T> {
        DataValue::UInt16(v)
    }
}

impl<T: IsParameterMask> From<u32> for DataValue<T> {
    fn from(v: u32) -> DataValue<T> {
        DataValue::UInt32(v)
    }
}

impl<T: IsParameterMask> From<u64> for DataValue<T> {
    fn from(v: u64) -> DataValue<T> {
        DataValue::UInt64(v)
    }
}

impl DataValueRaw {
    pub fn to_descriptive(
        &self,
        parameters: &[String],
    ) -> Result<DataValueDescriptive> {
        use DataValue::*;
        match *self {
            Boolean(v) => Ok(Boolean(v)),
            UInt8(v) => Ok(UInt8(v)),
            UInt16(v) => Ok(UInt16(v)),
            UInt32(v) => Ok(UInt32(v)),
            UInt64(v) => Ok(UInt64(v)),
            Int8(v) => Ok(Int8(v)),
            Int16(v) => Ok(Int16(v)),
            Int32(v) => Ok(Int32(v)),
            Int64(v) => Ok(Int64(v)),
            ParameterMask(v) => Ok(ParameterMask(
                v.deconstruct_parameter_mask(parameters)?,
            )),
        }
    }
}

impl DataValueDescriptive {
    pub fn to_raw(&self, parameters: &[String]) -> Result<DataValueRaw> {
        use DataValue::*;
        match *self {
            Boolean(ref v) => Ok(Boolean(*v)),
            UInt8(ref v) => Ok(UInt8(*v)),
            UInt16(ref v) => Ok(UInt16(*v)),
            UInt32(ref v) => Ok(UInt32(*v)),
            UInt64(ref v) => Ok(UInt64(*v)),
            Int8(ref v) => Ok(Int8(*v)),
            Int16(ref v) => Ok(Int16(*v)),
            Int32(ref v) => Ok(Int32(*v)),
            Int64(ref v) => Ok(Int64(*v)),
            ParameterMask(ref v) => Ok(ParameterMask(
                (&v).construct_parameter_mask(parameters)?,
            )),
        }
    }
}

impl<M: IsParameterMask> HasDataType for DataValue<M> {
    fn data_type(&self) -> DataType {
        use DataValue::*;

        match *self {
            Boolean(_) => DataType::Boolean,
            UInt8(_) => DataType::UInt8,
            UInt16(_) => DataType::UInt16,
            UInt32(_) => DataType::UInt32,
            UInt64(_) => DataType::UInt64,
            Int8(_) => DataType::Int8,
            Int16(_) => DataType::Int16,
            Int32(_) => DataType::Int32,
            Int64(_) => DataType::Int64,
            ParameterMask(_) => DataType::ParameterMask,
        }
    }
}

trait ConstructParameterMask {
    fn construct_parameter_mask(
        &self,
        parameters: &[String],
    ) -> Result<u64>;
}

impl<'a> ConstructParameterMask for &'a BTreeSet<String> {
    fn construct_parameter_mask(
        &self,
        parameters: &[String],
    ) -> Result<u64> {
        (&self.iter()
            .map(|v| v.to_string())
            .collect::<Vec<String>>()[..])
            .construct_parameter_mask(parameters)
    }
}

impl<'a> ConstructParameterMask for &'a [String] {
    fn construct_parameter_mask(
        &self,
        parameters: &[String],
    ) -> Result<u64> {
        let mut mask = 0u64;
        let mut unused = BTreeSet::new();
        for (i, parameter) in self.iter().enumerate() {
            if parameters.contains(&parameter) {
                if i >= 64 {
                    return Err(ErrorKind::ParameterOutOfMaskRange {
                        parameter: parameter.to_string(),
                        index: i,
                    }.into());
                } else {
                    mask |= 1u64 << i;
                }
            } else {
                unused.insert(parameter.to_string());
            }
        }
        if !unused.is_empty() {
            return Err(ErrorKind::ParametersNotInParametrSet {
                parameters: unused,
            }.into());
        }
        Ok(mask)
    }
}

trait DeconstructParameterMask {
    fn deconstruct_parameter_mask(
        &self,
        parameters: &[String],
    ) -> Result<BTreeSet<String>>;
}

impl DeconstructParameterMask for u64 {
    fn deconstruct_parameter_mask(
        &self,
        parameters: &[String],
    ) -> Result<BTreeSet<String>> {
        let mut mask = BTreeSet::new();
        for i in 0usize..64usize {
            if (self >> i) & 1u64 == 1u64 {
                if let Some(p) = parameters.get(i) {
                    mask.insert(p.to_string());
                } else {
                    return Err(ErrorKind::ParameterMaskOutOfRange {
                        index: i,
                        mask: *self,
                    }.into());
                }
            }
        }
        Ok(mask)
    }
}
