use failure::{Backtrace, Context, Fail};
use std::collections::BTreeSet;
use std::{fmt, result};
use {DataType, DataValueRaw};

pub type Result<T> = result::Result<T, Error>;

#[derive(Debug)]
pub struct Error {
    inner: Context<ErrorKind>,
}

impl Fail for Error {
    fn cause(&self) -> Option<&Fail> {
        self.inner.cause()
    }

    fn backtrace(&self) -> Option<&Backtrace> {
        self.inner.backtrace()
    }
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        fmt::Display::fmt(&self.inner, f)
    }
}

impl Error {
    pub fn kind(&self) -> ErrorKind {
        self.inner.get_context().clone()
    }
}

impl From<ErrorKind> for Error {
    fn from(kind: ErrorKind) -> Error {
        Error {
            inner: Context::new(kind),
        }
    }
}

impl From<Context<ErrorKind>> for Error {
    fn from(inner: Context<ErrorKind>) -> Error {
        Error { inner }
    }
}

#[derive(Clone, Eq, PartialEq, Debug, Fail)]
pub enum ErrorKind {
    #[fail(
        display = "value {:?} is not in range for type {:?}",
        value,
        datatype
    )]
    ValueOutOfRangeForType {
        value: DataValueRaw,
        datatype: DataType,
    },
    #[fail(
        display = "expected datatype compatible with {:?} found {:?}",
        expected,
        found
    )]
    IncompatibleDataType {
        expected: DataType,
        found: DataType,
    },
    #[fail(
        display = "parameter {:?} with index {} of mask range 64",
        parameter,
        index
    )]
    ParameterOutOfMaskRange {
        parameter: String,
        index: usize,
    },
    #[fail(
        display = "mask {:#b} has bit {} for non-existing parameters set",
        mask,
        index
    )]
    ParameterMaskOutOfRange { mask: u64, index: usize },
    #[fail(
        display = "parameters {:?} are not in parameter set", parameters
    )]
    ParametersNotInParametrSet { parameters: BTreeSet<String> },
}
